/*==============================================================================


   Purpose:    low level routines for SR8000


   Author:     Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de

		Modifications for SR8000: Helmut Heller, LRZ, Munich


   last mod:    23.08.2000
	   	26.9.2000
		This version is also for multithreaded COMPAS programs


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"

#include <stdlib.h>
#include <mach.h>

/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */


/* invalid entry-data if event is not available */
#define PCL_INVALID    0,{{{-1,-1,-1,-1,-1,-1,-1,-1,-1},{0,0,0,0,0,0,0,0,0}}}

/* Hitach SR8000 (8 counter + elapsed counter) */
static const PCL_event_descr_t event_table_sr8000[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,1,-1,-1,-1,-1,-1},{0,0,0,OP_ADD,0,0,0,0,0}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,1,-1,-1,-1,-1,-1,-1},{0,0,OP_ADD,0,0,0,0,0,0}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{1,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,0,0,0,0,0,0,0,0}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,1,-1,-1,-1,-1,-1,-1,-1},{0,OP_ADD,0,0,0,0,0,0,0}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,1,-1,-1},{0,0,0,0,0,0,OP_ADD,0,0}}}}, /* PCL_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,-1,1},{0,0,0,0,0,0,0,0,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,-1,-1,1,-1},{0,0,0,0,0,0,0,OP_ADD,0}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,1,-1,-1,-1,-1},{0,0,0,0,OP_ADD,0,0,0,0}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,-1,1,-1,-1,-1},{0,0,0,0,0,OP_ADD,0,0,0}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL */
#ifdef COMPASS
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,-1,1,1},{0,0,0,0,0,0,0,OP_FP_1|OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ}}}}, /* PCL_MFLOPS */
#else
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,-1,1,1,-1},{0,0,0,0,0,0,OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL,0}}}}, /* PCL_MFLOPS */
#endif
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,1,1,-1,-1},{0,0,0,0,0,OP_FP_1|OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,1,1,-1,-1,-1,-1},{0,0,0,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,0,0,0,0}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,1,-1,-1,1,-1},{0,0,0,0,OP_FP_2|OP_FP_STORE,0,0,OP_FP_2|OP_FP_DIV,0}}}} /* PCL_MEM_FP_RATIO */
};


/*                                                                            */
/*============================================================================*/
/* initialization                                                             */

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set event table */
  descr->event_table = event_table_sr8000;

  /* set allowed count mode */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter, /* NOT PCL_MAX_COUNTER */
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int i, rc;
#ifdef COMPASS
  int ip;
  pmc_data area;
#else
  uint64 counters[9];
#endif
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* read all counters */
#ifdef COMPASS
/*poption parallel */
  for(ip=0;ip<8;ip++){
    pmc_info( &(area.info[ip][0]) );
  }
#else
  pmc_info( counters );
#endif
 
  /* Extract selected events */
  for(i = 0; i < max_counter; ++i)
    if(IS_USED(used_mask, i))
    {
#ifdef COMPASS
    if(i==8)
    { /* We have to store all 8 counters individually to 
	    be able to compute the MAX later on.
	    We use hw_descr->elapsed_counter[ip] for this purpose. */
	for(ip=0;ip<8;ip++)
	{
	  hw_descr->elapsed_counter[ip] = area.info[ip][6];
	}
    }
    else
    {
	hw_descr->old_counters[i] = 0;
	for(ip=0;ip<8;ip++)
	{
	hw_descr->old_counters[i] += area.info[ip][i];
	}
    }
#else
      hw_descr->old_counters[i] = counters[i];
#endif
    }
  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter, /* NOT PCL_MAX_COUNTER */
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  int i, rc;
#ifdef COMPASS
  int ip;
  pmc_data area;
#else
  uint64 counters[9];
#endif
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* read all counters */
#ifdef COMPASS
/*poption parallel */
  for(ip=0;ip<8;ip++)
  {
    pmc_info( &(area.info[ip][0]) );
  }
#else
  pmc_info(counters);
#endif

  /* get the differences of old and current counters */
  for(i = 0; i < max_counter; ++i)
    if(IS_USED(counter_used_mask, i))
    {
#ifdef COMPASS
      if(i==8)
      /* PCL_ELAPSED_CYCLES */
      { /* Here we take the MAX() of all individual cycles to get s.th.
	      like the "wall clock time" */
	  counter_values[8] = area.info[0][6] - hw_descr->elapsed_counter[0];
	  for(ip=1;ip<8;ip++)
	  {
	      uint64 tmp;
	      tmp = area.info[ip][6] - hw_descr->elapsed_counter[ip];
	      if (tmp > counter_values[8]) counter_values[8] = tmp;
	  }
      }
	  else
      {
	  counter_values[i] = - hw_descr->old_counters[i];
	  for(ip=0;ip<8;ip++)
	  {
	    counter_values[i] += area.info[ip][i];
	  }
      }
#else
      if(i == 8)
        counter_values[i] = counters[6] - hw_descr->old_counters[6]; /* PCL_ELAPSED_CYCLES */
      else
        counter_values[i] = counters[i] - hw_descr->old_counters[i];
#endif
     }
#if 0
  for(i = 0; i < max_counter; ++i)
    if(IS_USED(counter_used_mask, i))
      printf("counter_values[%d] = %20.5f, max=%d\n", i, (double)counter_values[i], max_counter);
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */


int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
