/*==============================================================================


   Purpose:    small test program for low-level interface


   Author:     Rudolf Berrendorf
               University of Applied Sciences Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-rhein-sieg.de


   last mod:   28.09.2000



==============================================================================*/


#include <pcl.h>

#define N 1000


int
main(int argc, char **argv)

{
  int res, counter_index;
  PCL_BIT_MASK_TYPE counter_used_mask;
  PCL_DRIVER_COMMAND_TYPE commands[PCL_COUNTER_MAX];
  unsigned int count_mode;
  PCL_CNT_TYPE counter_values[PCL_COUNTER_MAX];
  int i;
  double a, b;
  PCL_DESCR_TYPE descr;


  counter_used_mask = 0;
  a = b = 1.0;

  /* allocate descriptor */
  if((res = PCLinit(&descr)) != PCL_SUCCESS)
    {
      printf("error on PCLinit (%d)\n", res);
      exit(1);
    }

  /* open driver */
  if((res = PCL_driver_open(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_open (%d)\n", res);
      exit(1);
    }


  /* start counter: we count FP instructions */
#if defined(ALPHA_CRAY)
  counter_index = 1;
  commands[counter_index] = 0x0a;

#elif defined(ALPHA_DEC)
  counter_index = 1;
  commands[counter_index] = 0x0a;

#elif defined(MIPS)
  counter_index = 1;
  commands[counter_index] = 16+5;

#elif defined(SPARC)
  /* not FP instructions but rather instructions */
  counter_index = 0;
  commands[counter_index] = (0x1 << 4);

#elif defined(POWERPC)
  counter_index = 1;
  commands[counter_index] = 15;

#elif defined(PENTIUM_LINUX)
  counter_index = 0;
  commands[counter_index] = 0xc1;

#elif defined(SR8000)
  counter_index = 7;
  commands[counter_index] = 1;

#else
#error "unsupported system"
#endif

  counter_used_mask = (0x1 << counter_index);
  count_mode = PCL_MODE_USER;
  if((res = PCL_driver_start(descr, PCL_COUNTER_MAX, counter_used_mask,
			     commands, count_mode)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_start (%d)\n", res);
      exit(1);
    }

  /* do some work */
  for(i = 0; i < N; ++i)
    a += (a+b);

  /* read counter */
  if((res = PCL_driver_read(descr, PCL_COUNTER_MAX, counter_used_mask,
			    counter_values)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_read (%d)\n", res);
      exit(1);
    }
  else
#if defined(SPARC)
    printf("%.0f instructions (dummy output=%2.0f)\n",
	   (double)counter_values[counter_index], a/a);
#else
    printf("%.0f floating point instructions, should be approx. %d "
	   "(dummy output=%2.0f)\n",
	   (double)counter_values[counter_index], 2 * N, a/a);
#endif

  /* stop counting */
  if((res = PCL_driver_stop(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_stop (%d)\n", res);
      exit(1);
    }

  /* close driver */
  if((res = PCL_driver_close(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_close (%d)\n", res);
      exit(1);
    }

  /* release descriptor */
  if((res = PCLexit(descr)) != PCL_SUCCESS)
    {
      printf("error on PCLexit_r (%d)\n", res);
      exit(1);
    }

  return 0;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
